#TODO: Implement dynamic icons!
#TODO: Finish Mocp line!


__all__ = (
    'Bar',
    'run',
    'write_initial_config'
)


import asyncio
import json
import os
import sys
import threading
import time
from copy import deepcopy

from .constants import (
    CLEAR_LINE,
    CONFIG_FILE,
    CSI,
    DEBUG,
    HIDE_CURSOR,
    UNHIDE_CURSOR,
)

from . import cli
from . import field_funcs
from . import utils
from .errors import *
from .field import Field, FieldPrecursor, FieldSpec
from .formatting import FormatStr, FmtStrStructure, FormatterFieldSig
from .namespaces import BarConfigSpec, BarSpec, FieldSpec
from ._types import (
    Args,
    ConsoleControlCode,
    FieldName,
    FieldOrder,
    Icon,
    JSONText,
    Kwargs,
    Line,
    PTY_Icon,
    PTY_Separator,
    Pattern,
    Separator,
    TTY_Icon,
    TTY_Separator,
)

from collections.abc import Container, Iterable, Iterator, Sequence
from os import PathLike
from typing import (
    IO,
    NoReturn,
    Required,
    Self,
    TypeVar
)

Bar = TypeVar('Bar')


class BarConfig(dict):
    '''
    Build and transport Bar configs between files, dicts and command line args.

    :param options: Optional :class:`_types.BarConfigSpec` parameters
        that override those of `defaults`
    :type options: :class:`_types.BarConfigSpec`

    :param defaults: Parameters to use by default,
        defaults to :attr:`Bar._default_params`
    :type defaults: :class:`dict`

    .. note:: `options` and `defaults` must be :class:`dict` instances
        of form :class:`_types.BarConfigSpec`

    '''

    def __init__(
        self,
        options: BarConfigSpec = {},
        defaults: BarConfigSpec = None,
    ) -> None:
        if defaults is None:
            self.defaults = Bar._default_params.copy()
        else:
            self.defaults = defaults.copy()
        self.options = options.copy()

        # This is why we subclass dict:
        self.update(self.defaults | self.options)
        self.file = None
        debug = self.options.pop('debug', None) or DEBUG

    def __repr__(self) -> str:
        cls = type(self).__name__
        file = self.file
        maybe_file = f"{file=}, " if file else ""
        params = {**self}
        return f"<{cls} {maybe_file}{params}>"

    @classmethod
    def from_file(
        cls,
        file: PathLike = None,
        *,
        defaults: BarConfigSpec = None,
        overrides: BarConfigSpec = {},
    ) -> Self:
        '''
        Return a new :class:`BarConfig` from a config file path.

        :param file: The filepath to the config file,
            defaults to ``'~/.mybar.json'``
        :type file: :class:`PathLike`

        :param defaults: The base :class:`_types.BarConfigSpec` dict whose
            params the new :class:`BarConfig` will override,
            defaults to :attr:`Bar._default_params`
        :type defaults: :class:`_types.BarConfigSpec`

        :param overrides: Additional param overrides to the config file
        :type overrides: :class:`_types.BarConfigSpec`

        :returns: A new :class:`BarConfig` instance
        :rtype: :class:`BarConfig`
        :raises: :exc:`OSError` for issues with accessing the file
        '''
        if defaults is None:
            defaults = Bar._default_params
        overrides = overrides.copy()

        if file is None:
            file = overrides.pop('config_file', CONFIG_FILE)
        absolute = os.path.abspath(os.path.expanduser(file))
        file_spec = {}
        try:
            file_spec, text = cls._read_file(absolute)
        except OSError as e:
            raise e.with_traceback(None)

        options = file_spec | overrides
        config = cls(options, defaults)
        config.file = absolute
        return config

    @classmethod
    def from_stdin(
        cls,
        write_new_file_dft: bool = True
    ) -> Self:
        '''Return a new :class:`BarConfig` using args from STDIN.
        Prompt the user before writing a new config file if one does
        not exist.

        :param write_new_file_dft: Write new files by default,
            defaults to ``True``
        :type write_new_file_dft: :class:`bool`

        :returns: A new :class:`BarConfig`
        :rtype: :class:`BarConfig`
        '''
        parser = cli.Parser()
        try:
            bar_options, command_options = parser.parse_args()
        except CLIUsageError as e:
            parser.error(e.msg)  # Shows usage

        # Handle options that alter the behavior of the command itself:
        if command_options:
            if 'dump_config' in command_options:
                indent = command_options.pop('dump_config', None)
                config = cls(bar_options)
                parser.quit(cls._as_json(config, indent=indent))

        # Handle missing config files:
        if 'config_file' not in bar_options:
            file = CONFIG_FILE
            if not os.path.exists(file):
                wants_to_write = cls.write_with_approval(overrides=bar_options)
                if not wants_to_write:
                    # Forget all this config file business.
                    # Our new user is in a hurry, so
                    # just give them what they need:
                    return cls(bar_options)

        else:
            file = bar_options.pop('config_file', None)

        try:
            # If 'config_file' is not in `overrides`,
            # it defaults to CONFIG_FILE.
            config = cls.from_file(file, overrides=bar_options)

        except FileNotFoundError as e:
            write_ok = cls.write_with_approval(file, overrides=bar_options)
            if not write_ok:
                parser.quit("Exiting...")
            config = cls.from_file(file, overrides=bar_options)

        except OSError as e:
            e.add_note("Exiting...")
            raise e from None

        except KeyboardInterrupt:
            parser.quit()

        return config

    @classmethod
    def write_with_approval(
        cls,
        file: PathLike = None,
        overrides: BarSpec = {},
    ) -> bool:
        '''
        Write a config file to `file`.
        Prompt the user for approval before writing.

        :param file: Write to this file, defaults to :obj:`CONFIG_FILE`
        :type file: :class:`PathLike`

        :returns: ``True`` if the file was written, ``False`` if not
        '''
        if file is None:
            file = CONFIG_FILE

        approved = cli.FileManager._get_write_new_approval(
            file,
            dft_choice=True
        )
        if approved:
            try:
                cls._write_file(file, overrides)
            except FileNotFoundError as e:
                try:
                    path = os.path.dirname(e.filename)
                    os.mkdir(path)
                    print(f"Made new directory {path!r}")
                    cls._write_file(file, overrides)
                except OSError as e:
                    raise e from None

            print(f"Wrote new config file to {file!r}")
            return True

        return False

    @staticmethod
    def _read_file(file: PathLike) -> tuple[BarConfigSpec, JSONText]:
        '''
        Read a given config file.
        Convert its JSON contents to a dict and return it along with the
        raw text of the file.

        :param file: The file to convert
        :type file: :class:`PathLike`

        :returns: The converted file and its raw text
        :rtype: tuple[:class:`_types.BarConfigSpec`, :class:`_types.JSONText`]
        '''
        absolute = os.path.abspath(os.path.expanduser(file))
        with open(absolute, 'r') as f:
            data = json.load(f)
            text = f.read()
        return data, text

    @classmethod
    def _write_file(
        cls,
        file: PathLike,
        spec: BarSpec = {},
        indent: int = 4,
        *,
        defaults: BarSpec = None
    ) -> None:
        '''
        Write :class:`BarConfig` params to a JSON file.

        :param file: The file to write to
        :type file: :class:`PathLike`

        :param spec: The :class:`_types.BarSpec` to write
        :type spec: :class:`_types.BarSpec`, optional

        :param indent: How many spaces to indent by, defaults to ``4``
        :type indent: :class:`int`

        :param defaults: Any default parameters that `spec` should override,
            defaults to :attr:`Bar._default_params`
        :type defaults: :class:`_types.BarSpec`
        '''
        un_pythoned = cls._remove_unserializable(spec, defaults=defaults)
        absolute = os.path.abspath(os.path.expanduser(file))
        if absolute == CONFIG_FILE and not os.path.exists(absolute):
            cli.FileManager._maybe_make_config_dir()
        with open(os.path.expanduser(absolute), 'w') as f:
            json.dump(un_pythoned, f, indent=indent, ) #separators=(',\n', ': '))

    @classmethod
    def _remove_unserializable(
        cls,
        spec: BarConfigSpec,
        *,
        defaults: BarSpec = None
    ) -> BarConfigSpec:
        '''
        Remove Python-specific API elements like functions which cannot
        be serialized for writing to files.

        :param spec: The :class:`_types.BarConfigSpec` to convert
        :type spec: :class:`_types.BarConfigSpec`

        :param defaults: Any default parameters that `spec` should override,
            defaults to :attr:`Bar._default_params`
        :type defaults: :class:`_types.BarSpec`
        '''
        if defaults is None:
            defaults = Bar._default_params.copy()

        newobj = spec.copy()
        newobj.pop('config_file', None)
        newobj = defaults | spec

        fields = newobj.pop('field_definitions', {})

        # Clean the dict of irrelevant Field implementation details:
        for name, field in fields.items():
            new = fields[name] = field.copy()
            for param in ('name', 'func', 'setup'):
                try:
                    del new[param]
                except KeyError:
                    pass

        if fields:
            newobj['field_definitions'] = fields

        return newobj

    @classmethod
    def _as_json(cls, spec: BarSpec = {}, indent: int = 4) -> JSONText:
        '''
        Return a :class:`BarConfig` as a string with JSON formatting.

        :param spec: The :class:`_types.BarSpec` to convert
        :type spec: :class:`_types.BarSpec`, optional

        :param indent: How many spaces to indent by, defaults to ``4``
        :type indent: :class:`int`
        '''
        cleaned = cls._remove_unserializable(spec)
        return json.dumps(cleaned, indent=indent)


class Bar:
    '''
    Create highly customizable status bars.

    :param fields: An iterable of default field names or :class:`Field` instances, defaults to ``None``
    :type fields: :class:`Iterable[Field | str]`

    :param template: A curly-brace format string with field names, defaults to ``None``
    :type template: :class:`formatting.FormatStr`

    :param separator: The field separator when `fields` is given, defaults to ``'|'``
    :type separator: :class:`_types.PTY_Separator` | :class:`_types.TTY_Separator`

    :param count: Only print the bar this many times, or never stop
        when ``None``, defaults to ``None``
    :type count: :class:`int`

    :param run_once: Whether the bar should print once and return, defaults to ``False``
    :type run_once: :class:`bool`

    :param refresh_rate: How often in seconds the bar automatically redraws itself, defaults to ``1.0``
    :type refresh_rate: :class:`float`

    :param align_to_seconds: Whether to synchronize redraws at the start of each new second (updates to the clock are shown accurately), defaults to ``True``
    :type align_to_seconds: :class:`bool`

    :param join_empty_fields: Whether to draw separators around fields with no content, defaults to ``False``
    :type join_empty_fields: :class:`bool`

    :param override_cooldown: Cooldown in seconds between handling sequential field overrides, defaults to ``1/60``
    :type override_cooldown: :class:`float`

    :param thread_cooldown: How long a field thread loop sleeps after checking if the bar is still running,
        defaults to ``1/8``.
        Between executions, unlike async fields, a threaded field sleeps for several iterations of
        `thread_cooldown` seconds that always add up to :attr:`Field.interval` seconds.
        Between sleeps, it checks if the bar has stopped.
        A shorter cooldown means more chances to check if the bar has stopped and
        a faster exit time.
    :type thread_cooldown: :class:`float`

    :param separators: A tuple of 2 strings that separate fields when `fields` is given.
        Note: The `separator` parameter sets both of these automatically.
        The first string is used in graphical (PTY) environments where support for Unicode is more likely.
        The second string is used in terminal (TTY) environments where only ASCII is supported.
        This enables the same :class:`Bar` instance to use the most optimal separator automatically.
    :type separators: tuple[:class:`_types.PTY_Separator`, :class:`_types.TTY_Separator`], optional

    :param stream: The bar's output stream, defaults to :attr:`sys.stdout`
    :type stream: :class:`IO`


    :raises: :exc:`errors.InvalidOutputStreamError` when `stream` does
        not implement the IO protocol
    :raises: :exc:`errors.IncompatibleArgsError` when
        neither `template` nor `fields` are given
    :raises: :exc:`errors.IncompatibleArgsError` when
        `template` is ``None`` but no `separator` or `separators` are given
    :raises: :exc:`TypeError` when `fields` is not iterable, or when `template` is not a string
    '''

    _default_field_order = [
        'hostname',
        'uptime',
        'cpu_usage',
        'cpu_temp',
        'mem_usage',
        'disk_usage',
        'battery',
        'net_stats',
        'datetime',
    ]

    _default_params = {
        'separator': '|',
        'refresh_rate': 1.0,
        'field_order': list(_default_field_order)
    }

    def __init__(
        self,
        template: FormatStr = None,
        fields: Iterable[Field | FieldName] = None,
        *,
        field_order: Iterable[FieldName] = None,
        separator: str = '|',
        refresh_rate: float = 1.0,
        stream: IO = sys.stdout,
        count: int = None,
        run_once: bool = False,
        align_to_seconds: bool = True,
        join_empty_fields: bool = False,
        override_cooldown: float = 1/8,
        thread_cooldown: float = 1/8,

        # Set this to use different seps for different output streams:
        separators: Sequence[PTY_Separator, TTY_Separator] = None,

        debug: bool = DEBUG,  # Not yet implemented!
    ) -> None:
        # Ensure the output stream has the required methods:
        self._check_stream(stream)
        self._stream = stream


        # Check all the required parameters.
        if fields is None:

            if template is None:
                raise IncompatibleArgsError(
                    f"Either a list of Fields 'fields' "
                    f"or a format string 'template' is required."
                )

            elif not isinstance(template, str):
                raise TypeError(
                    f"Format string 'template' must be a string, "
                    f"not {type(template)}"
                )

            # Good to use template:
            else:
                parsed = FmtStrStructure.from_str(template)
                parsed.validate_fields(Field._default_fields, True, True)
                # names = parsed.get_names()
                fields = parsed

        # Fall back to using fields.
        elif not hasattr(fields, '__iter__'):
            raise TypeError("The 'fields' argument must be iterable.")

        elif separator is None and separators is None:
            raise IncompatibleArgsError(
                "A separator is required when 'template' is None."
            )

        field_names, fields = self._normalize_fields(fields)

        # Preserve custom field order, enabling duplicates:
        if field_order is None:
            field_order = field_names

        self._fields = fields
        self._field_order = field_order
        self._buffers = dict.fromkeys(self._fields, '')

        self.template = template

        if separators is None:
            separators = (separator, separator)
        self._separators = separators

        # Whether empty fields are joined when template is None:
        # (True shows two separators together for every blank field.)
        self.join_empty_fields = join_empty_fields

        # How often in seconds the bar is regularly printed:
        self.refresh_rate = refresh_rate


        if count == 0 or count == 1:
            run_once = True
        self.count = count
        self._print_countdown = count
        # self._count_queue = asyncio.Queue()


        # Whether the bar should exit after printing once:
        self.run_once = run_once

        # Whether the bar is printed at the top of every clock second:
        self.align_to_seconds = align_to_seconds

        # Make a queue to which fields with overrides_refresh send updates.
        # Process only one item per refresh cycle to reduce flickering in a PTY.
        self._override_queue = asyncio.Queue(maxsize=1)

        # How long should the queue checker sleep before processing a new item?
        # (A longer cooldown means less flickering):
        self._override_cooldown = override_cooldown

        # Staggering a thread's refresh interval across several sleeps with the
        # length of this cooldown prevents it from blocking for the entire
        # interval when the bar stops.
        # (A shorter cooldown means faster exits):
        self._thread_cooldown = thread_cooldown

        # The dict mapping Field names to the threads running them.
        # The bar must join each before printing a line with `run_once`:
        self._threads = {}
        # self._threads = set()
        self._printer_thread = None
        self._printer_loop = None
        self._funcs = {}

        self._coros = {}
        self._timely_fields = []

        # Calling run() sets this Event. Unsetting it stops all fields:
        self._can_run = threading.Event()
        self._running = self._can_run.is_set

        # The bar's async event loop:
        # self._thread_loop = asyncio.new_event_loop()
        self._loop = asyncio.new_event_loop()

    def __contains__(self, other: FieldPrecursor) -> bool:
        if isinstance(other, str):
            weak_test = (other in self._field_order)
            return weak_test

        elif isinstance(other, Field):
            less_weak_test = (other.name in self._field_order)
            return less_weak_test

        else:
            return False

    def __eq__(self, other: Bar) -> bool:
        if not all(
            getattr(self, attr) == getattr(other, attr)
            for attr in (
                '_fields',
                '_field_order',
                '_separators',
                'join_empty_fields',
            )
        ):
            return False

        if self.template == other.template:
            return True
        return False

    def __iter__(self) -> Iterator:
        return iter(field for field in self._fields.values())

    def __len__(self) -> int:
        return len(self._field_order)

    def __repr__(self) -> str:
        names = self._field_order
        fields = utils.join_options(names, final_sep='', limit=3)
        cls = type(self).__name__
        return f"{cls}(fields=[{fields}])"

    @classmethod
    def from_config(
        cls,
        config: BarConfig,
        *,
        overrides: BarSpec = {},
        ignore_with: Pattern | tuple[Pattern] | None = '//'
    ) -> Self:
        '''
        Return a new :class:`Bar` from a :class:`BarConfig`
        or a dict of :class:`Bar` parameters.
        Ignore keys and list elements starting with `ignore_with`,
        which is ``'//'`` by default.
        If :param ignore_with: is ``None``, do not remove any values.

        :param config: The :class:`dict` to convert
        :type config: :class:`dict`

        :param overrides: Replace items in `config` with these parameters,
            defaults to ``{}``
        :type overrides: :class:`BarSpec`

        :param ignore_with: A pattern to ignore, defaults to ``'//'``
        :type ignore_with: :class:`_types.Pattern` | tuple[:class:`_types.Pattern`] | ``None``, optional

        :returns: A new :class:`Bar`
        :rtype: :class:`Bar`

        :raises: :exc:`errors.IncompatibleArgsError` when
            no `field_order` is defined
            or when no `template` is defined
        :raises: :exc:`errors.DefaultFieldNotFoundError` when either
            a field in `field_order` or
            a field in `template`
            cannot be found in :attr:`Field._default_fields`
        :raises: :exc:`errors.UndefinedFieldError` when
            a custom field name in `field_order` or
            a custom field name in `template`
            is not properly defined in `field_definitions`
        :raises: :exc:`errors.InvalidFieldSpecError` when
            a field definition is not of the form :class:`_types.FieldSpec`

        .. note:: `config` can be a regular :class:`dict`
        as long as it matches the form of :class:`_types.BarSpec`.

        '''
        if ignore_with is None:
            data = deepcopy(config)
        else:
            data = utils.scrub_comments(config, ignore_with)

        data.update(overrides)

        bar_params = cls._default_params | data
        field_order = bar_params.pop('field_order', None)
        field_icons = bar_params.pop('field_icons', {})  # From the CLI
        field_defs = bar_params.pop('field_definitions', {})

        if (template := bar_params.get('template')) is None:
            if field_order is None:
                raise IncompatibleArgsError(
                    "A bar format string 'template' is required "
                    "when field order list 'field_order' is undefined."
                )
        else:
            parsed = FmtStrStructure.from_str(template)
            parsed.validate_fields(Field._default_fields, True, True)
            if field_order is None:
                field_order = parsed.get_names()

        # Ensure icon assignments correspond to valid fields:
        for name in field_icons:
            if name not in field_order:
                deduped = dict.fromkeys(field_order)  # Preserve order
                expctd_from_icons = utils.join_options(deduped)
                exc = utils.make_error_message(
                    InvalidFieldError,
                    doing_what="parsing custom Field icons",
                    blame=f"{name!r}",
                    expected=f"a Field name from among {expctd_from_icons}",
                    epilogue=(
                        "Only assign icons to Fields that will be in the Bar."
                    )
                )
                raise exc from None

        # Gather Field parameters and instantiate new Fields:
        fields = {}
        expctd_name="the name of a default or properly defined `custom` Field"

        for name in field_order:
            field_params = field_defs.get(name)

            match field_params:
                case None:
                    # The field is strictly default.
                    if name in fields:
                        continue

                    try:
                        field = Field.from_default(name)
                    except DefaultFieldNotFoundError:
                        exc = utils.make_error_message(
                            DefaultFieldNotFoundError,
                            # doing_what="parsing 'field_order'",  # Only relevant to config file parsing
                            blame=f"{name!r}",
                            expected=expctd_name
                        )
                        raise exc from None

                case {'custom': True}:
                    # The field is custom, so it is only defined in
                    # 'field_definitions' and will not be found as a default.
                    name = field_params.pop('name', name)
                    del field_params['custom']
                    field = Field(**field_params, name=name)

                case {}:
                    # The field is a default overridden by the user.
                    # Are there custom icons from --icons in the CLI?
                    if field_icons and name in field_icons:
                        cust_icon = field_icons.pop(name)
                        field_params['icons'] = (cust_icon, cust_icon)

                    # When one icon is given, override the default icons:
                    elif 'icon' in field_params:
                        cust_icon = field_params['icon']
                        field_params['icons'] = (cust_icon, cust_icon)

                    try:
                        field = Field.from_default(name, overrides=field_params)
                    except DefaultFieldNotFoundError:
                        exc = utils.make_error_message(
                            UndefinedFieldError,
                            # doing_what="parsing 'field_order'",
                            blame=f"{name!r}",
                            expected=expctd_name,
                            epilogue=(
                                f"(In config files, remember to set "
                                f"`custom=true` "
                                f"for custom field definitions.)"
                            ),
                        )
                        raise exc from None

                case _:
                    # Edge cases.
                    exc = utils.make_error_message(
                        InvalidFieldSpecError,
                        # doing_what="parsing 'field_definitions'",
                        doing_what=f"parsing {name!r} definition",
                        details=(
                            f"Invalid Field specification: {field_params!r}",
                        ),
                        indent_level=1
                    )
                    raise exc from None

            fields[name] = field

        bar = cls(
            fields=tuple(fields.values()),
            field_order=field_order,
            **bar_params
        )
        return bar

    @classmethod
    def from_file(
        cls,
        file: PathLike = None,
        overrides: BarSpec = {}
    ) -> Self:
        '''
        Generate a new :class:`Bar` by reading a config file.

        :param file: The config file to read,
            defaults to :obj:`constants.CONFIG_FILE`
        :type file: :class:`PathLike`

        :param overrides: Replace items in `config` with these parameters,
            defaults to ``{}``
        :type overrides: :class:`BarSpec`

        :returns: A new :class:`Bar`
        :rtype: :class:`Bar`
        '''
        config = BarConfig.from_file(file)
        return cls.from_config(config, overrides=overrides)

    @classmethod
    def from_cli(cls) -> Self:
        '''
        Return a new :class:`Bar` using command line arguments.

        :returns: a new :class:`Bar` using command line arguments
        :rtype: :class:`Bar`
        '''
        config = BarConfig.from_stdin()
        return cls.from_config(config)

##    @classmethod
##    def as_generator(cls, 

    @property
    def clearline_char(self) -> str:
        '''
        A special character printed to TTY streams between refreshes.
        Its purpose is to clear characters left behind by longer lines.
        '''
        if self._stream is None:
            return None
        clearline = CLEAR_LINE if self._stream.isatty() else ''
        return clearline

    @property
    def fields(self) -> tuple[Field]:
        '''A tuple of the bar's Field objects.'''
        return tuple(self)

    @property
    def in_a_tty(self) -> bool:
        '''True if the bar was run from a terminal, otherwise False.'''
        if self._stream is None:
            return False
        return self._stream.isatty()

    @property
    def running(self) -> bool:
        return self._running()

    @property
    def separator(self) -> Separator:
        '''
        The field separator as determined by the output stream.
        It defaults to the TTY sep (self._separators[1]) if no stream is set.
        '''
        if self._stream is None:
            # Default to using the terminal separator:
            return self._separators[1]
        return self._separators[self._stream.isatty()]

    def append(self, field: FieldPrecursor) -> Self:
        '''
        Append a new Field to the bar.

        `field` will be converted to :class:`Field`,
        appended to the field list and joined to the end of the bar output.
        If :attr:`Bar.template` is defined, it will override the new field order.

        :param field: The field to append
        :type field: :class:`FieldPrecursor`

        :returns: The modified bar
        :rtype: :class:`Bar`
        '''
        (name,), normalized = self._normalize_fields((field,))
        self._fields.update(normalized)
        self._buffers[name] = ''
        self._field_order.append(name)
        return self

    def extend(self, fields: Iterable[FieldPrecursor]) -> Self:
        '''
        Append several new Fields to the bar.
        Field precursors in `fields` will be converted to :class:`Field`,
        appended to the field list and joined to the end of the bar output.
        If :attr:`Bar.template` is defined, it will override the new field order.
        the fields are displayed.

        :param field: The fields to append
        :type field: :class:`FieldPrecursor`

        :returns: The modified bar
        :rtype: :class:`Bar`
        '''
        #TODO: Consider Bar.template += FormatterFieldSig(field).as_string()
        names, normalized = self._normalize_fields(fields)
        self._fields.update(normalized)
        self._buffers.update(dict.fromkeys(names, ''))
        self._field_order.extend(names)
        return self

    @staticmethod
    def _check_stream(
        stream: IO,
        raise_on_fail: bool = True
    ) -> NoReturn | None:
        '''
        Check if an IO stream has proper methods for use by :class:`Bar`.
        If any methods are missing,
            return ``False`` if `raise_on_fail` is ``False``, or
            raise :exc:`InvalidBarError` if `raise_on_fail` is ``True``.
        Otherwise, return ``True``.

        :param stream: The IO stream object to test
        :type stream: :class:`IO`

        :param raise_on_fail: Raise exception if `stream` fails the check,
            defaults to ``True``
        :type raise_on_fail: :class:`bool`

        :raises: :exc:`InvalidOutputStreamError` if the stream fails the
            test and lacks required instance methods
        '''
        io_methods = ('write', 'flush', 'isatty')
        if not all(hasattr(stream, a) for a in io_methods):
            if not raise_on_fail:
                return False
            io_method_calls = [a + '()' for a in io_methods]
            joined = utils.join_options(io_method_calls, final_sep='and')
            raise InvalidOutputStreamError(
                f"Output stream {stream!r} needs {joined} methods."
            ) from None
        return True

    def _normalize_fields(
        self,
        fields: Iterable[FieldPrecursor],
    ) -> tuple[FieldOrder, dict[FieldName, Field]]:
        '''
        Convert :class:`Field` precursors to :class:`Field` instances.

        Take an iterable of field names, :class:`Field` instances, or
        :class:`FormatterFieldSig` tuples and convert them to
        corresponding default Fields.
        Return a dict mapping field names to Fields, along with a
        duplicate-preserving list of field names that were found.

        :param fields: An iterable of :class:`Field` primitives to convert
        :type fields: :class:`Iterable[FieldPrecursor]`

        :returns: A dict mapping field names to :class:`Field` instances
        :rtype: :class:`tuple[FieldOrder, dict[FieldName, Field]]`

        :raises: :exc:`errors.InvalidFieldError` when an element
            of `fields` is not a :class:`FieldPrecursor`
        '''
        normalized = {}
        names = []

##        if isinstance(fields, FmtStrStructure):
##            names = parsed.get_names()

        for field in fields:
            if isinstance(field, Field):
                new_field = field
                new_field._bar = self

            elif isinstance(field, str):
                new_field = Field.from_default(
                    name=field,
                    overrides={'bar': self},
                )

            elif isinstance(field, FormatterFieldSig):
                if field.name is None:
                    # Skip sigs that lack fields:
                    continue
                new_field = Field.from_default(
                    name=field.name,
                    overrides={'bar': self}
                )
                new_field._fmtsig = field.as_string()

            else:
                raise InvalidFieldError(f"Invalid field precursor: {field!r}")

            normalized[new_field.name] = new_field
            names.append(new_field.name)

        return names, normalized

    def line_generator(self) -> Iterator:
        '''
        Return a generator yielding status lines.
        This requires running the Fields in a separate thread,
        which has yet to be implemented.
        '''
        while self._running():
            if self._timely_fields:
                self._preload_timely_fields()
            line = self._make_one_line()
            yield line

    def current_line(self) -> Line:
        '''
        Return the most recently printed line.
        '''
        return self._make_one_line()

    def run(
        self,
        once: bool = None,
        stream: IO = None,
        *,
        bg: bool = False
    ) -> None:
        '''
        Run the bar in the specified output stream.
        Block until an exception is raised and exit smoothly.

        :param stream: The IO stream in which to run the bar,
            defaults to :attr:`Bar._stream`
        :type stream: :class:`IO`

        :param once: Run the bar only once,
            defaults to :attr:`Bar.run_once`
        :type once: :class:`bool`

        :param bg: (Not fully implemented)
            Run the bar in the background without printing,
            defaults to ``False``
            This is used for :func:`Bar.line_generator` and for testing.
        :type bg: :class:`bool`

        :returns: ``None``

        .. note::
            This method cannot be called within an async event loop.
        '''
        if self.count == 0:
            return

        try:
            if stream is not None:
                self._stream = stream
            if once is None:
                once = self.run_once
            else:
                self.run_once = once

            # Allow the bar to run repeatedly in the same interpreter:
            if self._loop.is_closed():
                self._loop = asyncio.new_event_loop()

            self._prepare_fields()
            self._can_run.set()
            for thread in tuple(self._threads.values()):
                thread.start()
            self._preload_timely_fields()

            if not once:
                self._start_printer()

            # This blocks:
            self._loop.run_until_complete(self._gather_coros())

            if once:
                # Wait for threads to finish to get a full line, then print.
                while self._threads:
                    time.sleep(self._thread_cooldown)
                self._print_one_line()
                self._can_run.clear()

            # Prevent the bar from exiting when there are no coroutines:
            while self._running():
                time.sleep(self._thread_cooldown)

        except KeyboardInterrupt:
            pass

        finally:
            self._shutdown()

    def _prepare_fields(self) -> None:
        overriding = False
        for field in self:
            if field.overrides_refresh:
                overriding = True

            if field.threaded:
                thread = field.make_thread(bar=self, run_once=self.run_once)
                self._threads[thread.name] = thread
            elif field.timely:
                self._timely_fields.append(field)
            else:
                self._coros[field.name] = field.run(
                    bar=self, once=self.run_once
                )

##        if overriding:  # Currently disabled!
##            self._coros.append(self._handle_overrides())

    async def _gather_coros(self) -> None:
        '''
        Run :class:`Bar` coroutines in parallel.
        '''
        await asyncio.gather(*self._coros.values())

    def _preload_timely_fields(self) -> None:
        '''
        Update the bar buffers for Fields with :attr:`Field.timely`.
        This is used most often right before printing a line.
        '''
        for f in self._timely_fields:
            if f.is_async:
                result = self._loop.run_until_complete(
                    f._func(*f.args, **f.kwargs)
                )
            else:
                result = f._func(*f.args, **f.kwargs)

            if f.template is not None:
                contents = f.template.format(result, icon=f.icon)
            else:
                if f.always_show_icon or result:
                    contents = f.icon + result
                else:
                    contents = result
            self._buffers[f.name] = contents

    def _start_printer(self, end: str = '\r') -> None:
        '''
        Make a thread in which to run
            :meth:`Bar._threaded_continuous_line_printer` and start it.
        '''
        thread_name = 'PRINTER'
        self._printer_thread = threading.Thread(
            target=self._threaded_continuous_line_printer,
            name=thread_name,
            args=(end,)
        )
        self._threads[thread_name] = self._printer_thread
        self._printer_thread.start()

    def _threaded_continuous_line_printer(self, end: str = '\r') -> None:
        '''
        The bar's primary line-printing mechanism.
        Fields with the ``timely`` attribute get run here.
        Other fields are responsible for sending data to bar buffers.
        This only writes using the current buffer contents.

        :param end: The string appended to the end of each line
        :type end: :class:`str`
        '''
        # Make an event loop for this thread:
        self._printer_loop = asyncio.new_event_loop()

        # Flushing the buffer before writing to it fixes poor i3bar
        # alignment.
        self._stream.flush()

        if self.in_a_tty:
            beginning = self.clearline_char + end
            self._stream.write(CSI + HIDE_CURSOR)
        else:
            beginning = self.clearline_char

        # Print something right away just so that the bar is not empty,
        # but not if the print count matters:
        if not self.count:
            line = self._make_one_line()
            self._stream.write(beginning + line + end)
            self._stream.flush()

        using_format_str = (self.template is not None)
        sep = self.separator
        clock = time.monotonic
        step = self._thread_cooldown
        needed = round(self.refresh_rate / step)
        count = 0
        first_cycle = True

        if self.align_to_seconds:
            # Sleep until the beginning of the next second.
            clock = time.time
            time.sleep(1 - (clock() % 1))

        start_time = clock()
        while self._running():

            # Sleep until the next refresh cycle in little steps to
            # check if the bar has stopped.
            if count < needed and not first_cycle:
                count += 1

                # Latency from nonzero execution times causes drift,
                # where sleeps become out-of-sync and the bar skips
                # field updates.
                # This is especially noticeable in fields that update
                # routinely, such as the time.

                # To negate drift, when sleeping until the beginning of
                # the next cycle, we must also compensate for latency.
                time.sleep(
                    step - (
                        # Get the current latency, which can vary:
                        clock() % step
                        # (clock() - start_time) % step To preserve offset
                    )
                )
                continue

            count = 0
            if first_cycle:
                first_cycle = False

            # Run time-sensitive fields right away:
            for f in self._timely_fields:
                if f.is_async:
                    result = self._printer_loop.run_until_complete(
                        f._func(*f.args, **f.kwargs)
                    )
                else:
                    result = f._func(*f.args, **f.kwargs)

                if f.template is not None:
                    contents = f.template.format(result, icon=f.icon)
                else:
                    if f.always_show_icon or result:
                        contents = f.icon + result
                    else:
                        contents = result
                self._buffers[f.name] = contents

            if using_format_str:
                line = self.template.format_map(self._buffers)
            else:
                line = sep.join(
                    buf for field in self._field_order
                        if (buf := self._buffers[field])
                        or self.join_empty_fields
                )

            self._stream.write(beginning + line + end)
            self._stream.flush()

            if self.count:
                self._print_countdown -= 1
                if self._print_countdown == 0:
                    self._can_run.clear()
                    self._printer_loop.stop()
                    self._printer_loop.close()
                    return

        self._printer_loop.stop()
        self._printer_loop.close()

    def _shutdown(self) -> None:
        '''
        Notify fields that the bar has stopped and join threads.
        Also print a newline and unhide the cursor if the bar was
        running in a terminal.
        '''
        self._can_run.clear()

        threads = tuple(self._threads.items())
        for thread_name, thread in threads:
            thread.join()
            self._threads.pop(thread_name)

        if self.in_a_tty:
            self._stream.write('\n')
            self._stream.write(CSI + UNHIDE_CURSOR)

    async def _handle_overrides(self, end: str = '\r') -> None:
        '''
        Print a line when fields with overrides_refresh send new data.

        :param end: The string appended to the end of each line
        :type end: :class:`str`
        '''
        sep = self.separator
        using_format_str = (self.template is not None)

        if self.in_a_tty:
            beginning = self.clearline_char + end
        else:
            beginning = self.clearline_char

        start_time = time.time()
        while self._running():

            try:
                # Wait until a field with overrides_refresh sends new
                # data to be printed:
                field, contents = await self._override_queue.get()
                # if DEBUG:
                    # logger.debug(f"handler: {field} {time.time() - start_time}")

            except RuntimeError:
                # asyncio raises RuntimeError if the event loop closes
                # while queue.get() is waiting for a value.
                return

            if using_format_str:
                line = self.template.format_map(self._buffers)
            else:
                line = sep.join(
                    buf for field in self._field_order
                        if (buf := self._buffers[field])
                        or self.join_empty_fields
                )

            self._stream.write(beginning + line + end)
            self._stream.flush()
            await asyncio.sleep(self._override_cooldown)

    def _make_one_line(self) -> Line:
        '''Make a line using the bar's field buffers.
        This method is not meant to be called from within a loop.
        '''
        if self.template is not None:
            line = self.template.format_map(self._buffers)
        else:
            line = self.separator.join(
                buf for field in self._field_order
                    if (buf := self._buffers[field])
                    or self.join_empty_fields
            )
        return line

    def _print_one_line(self, end: str = '\r') -> None:
        '''Print a line to the buffer stream only once.
        This method is not meant to be called from within a loop.

        :param end: The string appended to the end of each line
        :type end: :class:`str`
        '''
        if self.in_a_tty:
            beginning = self.clearline_char + end
        else:
            beginning = self.clearline_char

        if self._timely_fields:
            self._preload_timely_fields()
        self._stream.write(beginning + self._make_one_line() + end)

        self._stream.flush()


def run(once: bool = False, file: PathLike = None) -> NoReturn | None:
    '''
    Generate a new :class:`Bar` from a config file and run it in STDOUT.
    Ask to write the file if it doesn't exist.

    :param once: Print the bar only once, defaults to ``False``
    :type once: :class:`bool`

    :param file: The config file to source,
        defaults to :obj:`constants.CONFIG_FILE`
    :type file: :class:`PathLike`
    '''
##    import __main__ as possibly_repl
##    if not hasattr(possibly_repl, '__file__'):
##        # User is in a REPL
    try:

        try:
            bar = Bar.from_file(file)
        except FileNotFoundError as e:
            _p_ = __package__
            msg = (
                f"Try running `{_p_}.write_initial_config()`"
                f" if this is your first time using {_p_}."
            )
            e.add_note(msg)
            raise e from None

        bar.run(once=once)

    except KeyboardInterrupt:
        print()
        return


def write_initial_config() -> None:
    BarConfig.write_with_approval(CONFIG_FILE)




field_defs = {

    'hostname': {
        'name': 'hostname',
        'func': field_funcs.get_hostname,
        'run_once': True
    },

    'host': {
        'name': 'host',
        'func': field_funcs.get_host,
        'kwargs': {
            'fmt': '{nodename}',
        },
        'run_once': True
    },

    'uptime': {
        'name': 'uptime',
        'func': field_funcs.get_uptime,
        'setup': _setups.setup_uptime,
        'kwargs': {
            'fmt': '{days}d:{hours}h:{mins}m',
            'sep': ':'
        },
        'icons': [' ', 'Up '],
    },

    'cpu_usage': {
        'name': 'cpu_usage',
        'func': field_funcs.get_cpu_usage,
        'interval': 2,
        'threaded': True,
        'icons': [' ', 'CPU '],
    },

    'cpu_temp': {
        'name': 'cpu_temp',
        'func': field_funcs.get_cpu_temp,
        'interval': 2,
        'threaded': True,
        'icons': [' ', ''],
    },

    'mem_usage': {
        'name': 'mem_usage',
        'func': field_funcs.get_mem_usage,
        'interval': 2,
        'icons': [' ', 'Mem '],
    },

    'disk_usage': {
        'name': 'disk_usage',
        'func': field_funcs.get_disk_usage,
        'interval': 4,
        'icons': [' ', '/:'],
    },

    'battery': {
        'name': 'battery',
        'func': field_funcs.get_battery_info,
        'icons': [' ', 'Bat '],
    },

    'net_stats': {
        'name': 'net_stats',
        'func': field_funcs.get_net_stats,
        'interval': 4,
        'threaded': True,
        'icons': [' ', ''],
    },

    'datetime': {
        'name': 'datetime',
        # 'func': field_funcs.precision_datetime,
        'func': field_funcs.get_datetime,
        'align_to_seconds': True,
        'timely': True,
        'kwargs': {
            'fmt': "%Y-%m-%d %H:%M:%S"
        },
    }

}

field_order = [
    'hostname',
    'uptime',
    'cpu_usage',
    'cpu_temp',
    'mem_usage',
    'disk_usage',
    'battery',
    'net_stats',
    'datetime',
]

bar_params = {
    'separator': '|',
    'refresh_rate': 1.0,
    'field_order': list(_default_field_order)
}




#TODO: Implement dynamic icons!
#TODO: Finish Mocp line!


__all__ = (
    'Field',
)


import asyncio
import threading
import time

from . import field_funcs
from . import _setups
from . import utils
from .constants import DEBUG
from .errors import *
from .formatting import FormatterFieldSig
from .namespaces import FieldSpec
from ._types import (
    FieldName,
    Icon,
    PTY_Icon,
    TTY_Icon,
    FormatStr,
    Args,
    Kwargs,
)

from collections.abc import Callable, Sequence
from typing import (
    NamedTuple,
    NoReturn,
    ParamSpec,
    Required,
    TypeAlias,
    TypeVar
)

Bar_T = TypeVar('Bar')
Field = TypeVar('Field')
P = ParamSpec('P')


class Field:
    '''
    Continuously generates and formats one bit of information in a :class:`Bar`.
    Pre-existing default Fields can be looked up by name.

    :param name: A unique identifier for the new field, defaults to `func.`:attr:`__name__`
    :type name: :class:`str`

    :param func: The Python function to run at every `interval` if no `constant_output` is set
    :type func: :class:`Callable[[*Args, **Kwargs], str]`

    :param icon: The field icon, defaults to ``''``.
        Placed before field contents or in place
        of ``{icon}`` in `template`, if provided
    :type icon: :class:`str`

    :param template: A curly-brace format string.
        This parameter is **required** if `icon` is ``None``.
        Valid placeholders:
            - ``{icon}`` references `icon`
            - ``{}`` references field contents
        Example:
            When the field's current contents are ``'69F'`` and its icon is ``'TEMP'``,
            ``template='[{icon}]: {}'`` shows as ``'[TEMP]: 69F'``
    :type template: :class:`_types.FormatStr`

    :param interval: How often in seconds field contents are updated, defaults to ``1.0``
    :type interval: :class:`float`

    :param align_to_seconds: Update contents at the start of each second, defaults to ``False``
    :type align_to_seconds: :class:`bool`

    :param timely: Run the field as soon as possible after every refresh, defaults to ``False``
    :type timely: :class:`bool`

    :param overrides_refresh: Updates to this field re-print the bar between refreshes, defaults to ``False``
    :type overrides_refresh: :class:`bool`

    :param threaded: Run this field in a separate thread, defaults to ``False``
    :type threaded: :class:`bool`

    :param always_show_icon: Show icons even when contents are empty, defaults to ``False``
    :type always_show_icon: :class:`bool`

    :param run_once: Permanently set contents by running `func` only once, defaults to ``False``
    :type run_once: :class:`bool`

    :param constant_output: Permanently set contents instead of running a function
    :type constant_output: :class:`str`, optional

    :param bar: Attach the :class:`Field` to this :class:`Bar`
    :type bar: :class:`Bar`, optional

    :param args: Positional args passed to `func`
    :type args: :class:`Args`, optional

    :param kwargs: Keyword args passed to `func`
    :type kwargs: :class:`Kwargs`, optional

    :param setup: A special callback that updates `kwargs` with static data that `func` would otherwise have to evaluate every time it runs
    :type setup: :class:`Callable[P, Kwargs]`, optional

    :param icons: A pair of icons used in different cases.
        Note: The `icon` parameter sets both of these automatically.
        The first string is intended for graphical (PTY) environments where support for Unicode is more likely.
        The second string is intended for terminal (TTY) environments where only ASCII is supported.
        This enables the same :class:`Field` instance to use the most optimal icon automatically.
    :type icons: tuple[:class:`_types.PTY_Icon`, :class:`_types.TTY_Icon`], optional


    :raises: :exc:`errors.IncompatibleArgsError` when
        neither `func` nor `constant_output` are given
    :raises: :exc:`errors.IncompatibleArgsError` when
        neither `icon` nor `template` are given
    :raises: :exc:`TypeError` when `func` is not callable
    :raises: :exc:`TypeError` when `setup`, if given, is not callable
    '''
    _default_fields = {

        'hostname': {
            'name': 'hostname',
            'func': field_funcs.get_hostname,
            'run_once': True
        },

        'host': {
            'name': 'host',
            'func': field_funcs.get_host,
            'kwargs': {
                'fmt': "{nodename}",
            },
            'run_once': True
        },

        'uptime': {
            'name': 'uptime',
            'func': field_funcs.get_uptime,
            'kwargs': {
                'fmt': '{days}d:{hours}h:{mins}m',
                'dynamic': True,
                'sep': ':',
            },
            'setup': _setups.setup_uptime,
            'timely': True,
            'align_to_seconds': True,
            'icon': 'Up ',
        },

        'cpu_usage': {
            'name': 'cpu_usage',
            'func': field_funcs.get_cpu_usage,
            'kwargs': {
                'fmt': "{:02.0f}%",
            },
            'interval': 5,
            'threaded': True,
            'icon': 'CPU ',
        },

        'cpu_temp': {
            'name': 'cpu_temp',
            'func': field_funcs.get_cpu_temp,
            'kwargs': {
                'fmt': "{temp:02.0f}{scale}",
            },
            'interval': 5,
            'threaded': True,
        },

        'mem_usage': {
            'name': 'mem_usage',
            'func': field_funcs.get_mem_usage,
            'kwargs': {
                'fmt': "{used:.1f}{unit}",
                'unit': 'G',
            },
            'interval': 5,
            'icon': 'Mem ',
        },

        'disk_usage': {
            'name': 'disk_usage',
            'func': field_funcs.get_disk_usage,
            'kwargs': {
                'fmt': "{free:.1f}{unit}",
                'path': '/',
                'unit': 'G',
            },
            'interval': 5,
            'icon': '/:',
        },

        'battery': {
            'name': 'battery',
            'kwargs': {
                'fmt': "{pct:02.0f}{state}",
            },
            'func': field_funcs.get_battery_info,
            'threaded': True,
            'icon': 'Bat ',
        },

        'net_stats': {
            'name': 'net_stats',
            'func': field_funcs.get_net_stats,
            'kwargs': {
                # 'device': None,
                'fmt': "{name}",
                'nm': True,
                'nm_filt': None,
                'default': "",
            },
            'interval': 5,
        },

        'datetime': {
            'name': 'datetime',
            'func': field_funcs.get_datetime,
            'kwargs': {
                'fmt': "%Y-%m-%d %H:%M:%S",
            },
            'align_to_seconds': True,
            'timely': True,
        }

    }

    def __init__(
        self,
        *,
        name: FieldName = None,
        func: Callable[P, str] = None,
        icon: str = '',
        template: FormatStr = None,
        interval: float = 1.0,
        align_to_seconds: bool = False,
        timely: bool = False,
        overrides_refresh: bool = False,
        threaded: bool = False,
        always_show_icon: bool = False,
        run_once: bool = False,
        constant_output: str = None,
        bar: Bar_T = None,
        args: Args = None,
        kwargs: Kwargs = None,
        setup: Callable[P, P.kwargs] = None,

        # Set this to use different icons for different output streams:
        icons: Sequence[PTY_Icon, TTY_Icon] = None,
    ) -> None:

        if constant_output is None:
            #NOTE: This will change when dynamic icons and templates are implemented.
            if func is None:
                raise IncompatibleArgsError(
                    f"Either a function that returns a string or "
                    f"a constant output string is required."
                )
            if not callable(func):
                raise TypeError(
                    f"Type of 'func' must be callable, not {type(func)}"
                )

        if name is None and callable(func):
            name = func.__name__
        self.name = name

        self._func = func
        self.args = () if args is None else args
        self.kwargs = {} if kwargs is None else kwargs

        if setup is not None and not callable(setup):
            # `setup` was given but is of the wrong type.
            raise TypeError(
                f"Type of 'setup' must be callable, not {type(setup)}"
            )
        self._setupfunc = setup

        if icons is None:
            icons = (icon, icon)
        self._icons = icons

        if template is None and icons is None:
            raise IncompatibleArgsError(
                "An icon is required when `template` is None."
            )
        self.template = template

        self.is_async = asyncio.iscoroutinefunction(func)

        if self.is_async or threaded or timely:
            self._callback = func
        else:
            # Wrap synchronous functions if they aren't special:
            self._callback = self._asyncify

        self._bar = bar

        self.timely = utils.str_to_bool(timely)
        self.align_to_seconds = utils.str_to_bool(align_to_seconds)
        self.always_show_icon = utils.str_to_bool(always_show_icon)
        self._buffer = None
        self.constant_output = constant_output
        self.interval = float(interval)
        self.overrides_refresh = utils.str_to_bool(overrides_refresh)
        self.run_once = utils.str_to_bool(run_once)

        self.threaded = utils.str_to_bool(threaded)
        self._do_setup()

    def __repr__(self) -> str:
        cls = type(self).__name__
        name = self.name
        # attrs = utils.join_options(...)
        return f"{cls}({name=})"

    def __eq__(self, other: Field) -> bool:
        if not all(
            getattr(self, attr) == getattr(other, attr)
            for attr in (
                'align_to_seconds',
                'always_show_icon',
                'constant_output',
                'template',
                '_icons',
            )
        ):
            return False

        if self._func is None:
            if self.constant_output == other.constant_output:
                return True
        elif (
            self._func is other._func
            and self.args == other.args
            and self.kwargs == other.kwargs
        ):
            return True
        return False

    @classmethod
    def from_default(
        cls: Field,
        name: str,
        *,
        overrides: FieldSpec = {},
        source: dict[FieldName, FieldSpec] = None,
        fmt_sig: FormatStr = None
    ) -> Field:
        '''Quickly get a default Field and customize its parameters.

        :param name: Name of the default :class:`Field` to access or customize
        :type name: :class:`str`

        :param overrides: Custom parameters that override those of the default Field
        :type overrides: :class:`_types.FieldSpec`, optional

        :param source: The :class:`dict` in which to look up default fields,
            defaults to :attr:`Field._default_fields`
        :type source: dict[:class:`_types.FieldName`, :class:`_types.FieldSpec`]

        :returns: A new :class:`Field`
        :rtype: :class:`Field`
        :raises: :exc:`errors.DefaultFieldNotFoundError` when `source` does not contain `name`
        '''
        if source is None:
            source = cls._default_fields

        try:
            default: FieldSpec = source[name].copy()
        except KeyError:
            raise DefaultFieldNotFoundError(
                f"{name!r} is not the name of a default Field."
            ) from None

        if 'kwargs' in overrides and 'kwargs' in default:
            # default['kwargs'].update(overrides['kwargs'])
            # default['kwargs'] |= overrides['kwargs']
            overrides['kwargs'] = default['kwargs'] | overrides['kwargs']

        spec = default | overrides
        field = cls(**spec)
        field._fmt_sig = fmt_sig
        return field

    @classmethod
    def from_format_string(cls, fmt: FormatStr) -> Field:
        '''
        Get a :class:`Field` from a curly-brace field in a format string.

        :param template: The format string to convert
        :type template: :class`FormatStr`
        '''
        sig = FormatterFieldSig.from_str(fmt)
        field = cls.from_default(sig.name)
        field._fmtsig = fmt
        return field

    @property
    def icon(self) -> str:
        '''The field icon as determined by the output stream of its bar.
        It defaults to the TTY icon (:attr:`self._icons[1]`) if no bar is set.
        '''
        if self._bar is None:
            return self._icons[1]  # Default to using the terminal icon.
        return self._icons[self._bar._stream.isatty()]

    async def _asyncify(self, *args, **kwargs) -> str:
        '''Wrap a synchronous function in a coroutine for simplicity.'''
        return self._func(*args, **kwargs)

    @staticmethod
    def _format_contents(
        text: str,
        icon: str,
        template: FormatStr = None,
        always_show_icon: bool = False
    ) -> str:
        '''A helper function that formats field contents.'''
        if template is None:
            if always_show_icon or text:
                return icon + text
            else:
                return text
        else:
            return template.format(text, icon=icon)

    def _auto_format(self, text: str) -> 'Contents':
        '''Non-staticmethod _format_contents...'''
        if self.template is None:
            if self.always_show_icon or text:
                return self.icon + text
            else:
                return text
        else:
            return self.template.format(text, icon=self.icon)

    def as_generator(self, once: bool = False) :
        yield self._func

    def _do_setup(self, args: Args = None, kwargs: Kwargs = None) -> None:
        # Use the pre-defined _setupfunc() to gather constant variables
        # for func() which might only be evaluated at runtime:
        if self._setupfunc is None:
            return

        if args is None:
            args = self.args
        if kwargs is None:
            kwargs = self.kwargs

        try:
            if asyncio.iscoroutinefunction(self._setupfunc):
                setupvars = asyncio.get_event_loop().run_until_complete(
                    self._setupfunc(*self.args, **self.kwargs)
                )
            else:
                setupvars = self._setupfunc(*self.args, **self.kwargs)

        # If _setupfunc raises FailedSetup with a backup value,
        # use it as the field's new constant_output:
        except FailedSetup as e:
            backup = e.backup
            self.constant_output = self._auto_format(backup)
            return

        # On success, give new values to kwargs to pass to _func().
        # return setupvars
        kwargs['setupvars'] = setupvars


##    # For threaded fields, still run continuously!!!

    def sync_run(self, once: bool = None) -> 'Contents':
        # Do not run fields which have a constant output;
        # only set their bar buffer.
        if self.constant_output is not None:
            return self._auto_format(self.constant_output)

        if self.is_async:
            result = asyncio.get_event_loop().run_until_complete(self._func(*self.args, **self.kwargs))
        else:
            result = self._func(*self.args, **self.kwargs)
        contents = self._auto_format(result)

        return contents

    def gen_run(self, once: bool = None) -> 'Contents':
        # Do not run fields which have a constant output;
        # only set their bar buffer.
        if self.constant_output is not None:
            return self._auto_format(self.constant_output)

        while True:
            if self.is_async:
                result = self._func(*self.args, **self.kwargs)
            else:
                result = asyncio.get_event_loop().run_until_complete(self._func(*self.args, **self.kwargs))
            contents = self._auto_format(result)

            yield contents

    async def run(self, bar: Bar_T, once: bool) -> None:
        '''
        Run an async :class:`Field` callback and send its output to a
        status bar.

        :param bar: Send results to this status bar
        :type bar: :class:`Bar`

        :param once: Run the Field function once
        :type once: :class:`bool`
        '''
        self._check_bar(bar)
        # Do not run fields which have a constant output;
        # only set their bar buffer.
        if self.constant_output is not None:
            contents = self._format_contents(
                self.constant_output,
                self.icon,
                self.template,
                self.always_show_icon
            )
            bar._buffers[self.name] = contents
            return

        func = self._callback
        running = bar._can_run.is_set
        clock = time.monotonic
        using_format_str = (self.template is not None)
        last_val = None

        # Use the pre-defined _setupfunc() to gather constant variables
        # for func() which might only be evaluated at runtime:
        if self._setupfunc is not None:
            try:
                if asyncio.iscoroutinefunction(self._setupfunc):
                    setupvars = (
                        await self._setupfunc(*self.args, **self.kwargs)
                    )
                else:
                    setupvars = self._setupfunc(*self.args, **self.kwargs)

            # If _setupfunc raises FailedSetup with a backup value,
            # use it as the field's new constant_output and update the
            # bar's buffer:
            except FailedSetup as e:
                backup = e.backup
                contents = self._format_contents(
                    backup,
                    self.icon,
                    self.template,
                    self.always_show_icon
                )
                self.constant_output = contents
                bar._buffers[self.name] = str(contents)
                return

            # On success, give new values to kwargs to pass to func().
            self.kwargs['setupvars'] = setupvars

        # Run at least once at the start to ensure bars have contents:
        result = await func(*self.args, **self.kwargs)
        last_val = result
        contents = self._auto_format(result)
        bar._buffers[self.name] = contents

        if self.run_once or once:
            return

        if self.align_to_seconds:
            # Sleep until the beginning of the next second.
            clock = time.time
            await asyncio.sleep(1 - (clock() % 1))
        start_time = clock()

        # The main loop:
        while running():

            if bar.count:
                # Stop before running a cycle that could last too long:
                if (
                    bar._print_countdown == 0
                    or self.interval > (
                        bar._print_countdown * bar.refresh_rate 
                    )
                ):
                    bar._coros.pop(self.name, None)
                    return

            result = await func(*self.args, **self.kwargs)
            # Latency from nonzero execution times causes drift, where
            # sleeps become out-of-sync and the bar skips field updates.
            # This is especially noticeable in fields that update
            # routinely, such as the time.

            # To negate drift, when sleeping until the beginning of the
            # next cycle, we must also compensate for latency.
            await asyncio.sleep(
                # Time until next refresh:
                self.interval - (
                    # Get the current latency, which can vary:
                    clock() % self.interval
                    # (clock() - start_time) % self.interval  # Preserve offset
                )
            )

            if result == last_val:
                continue
            last_val = result

            if using_format_str:
                contents = self.template.format(result, icon=self.icon)
            else:
                if self.always_show_icon or result:
                    contents = self.icon + result
                else:
                    contents = result

            bar._buffers[self.name] = contents

            # Send new field contents to the bar's override queue and print a
            # new line between refresh cycles.
            if self.overrides_refresh:
                try:
                    bar._override_queue.put_nowait((self.name, contents))

                except asyncio.QueueFull:
                    # Since the bar buffer was just updated, do nothing if the
                    # queue is full. The update may still show while the queue
                    # handles the current override.
                    # If not, the line will update at the next refresh cycle.
                    pass

    def run_threaded(self, bar: Bar_T, once: bool) -> None:
        '''
        Run a blocking :class:`Field` func and send its output to a
        status bar.

        :param bar: Send results to this status bar
        :type bar: :class:`Bar`

        :param once: Run the Field function once
        :type once: :class:`bool`
        '''
        self._check_bar(bar)

        # Do not run fields which have a constant output;
        # only set their bar buffer.
        if self.constant_output is not None:
            contents = self._format_contents(
                self.constant_output,
                self.icon,
                self.template,
                self.always_show_icon
            )
            bar._buffers[self.name] = contents
            return

        func = self._callback
        running = bar._can_run.is_set
        clock = time.monotonic
        using_format_str = (self.template is not None)
        last_val = None

        # If the field's callback is asynchronous,
        # it must be run in a new event loop.
        local_loop = asyncio.new_event_loop()

        # Use the pre-defined _setupfunc() to gather constant variables
        # for func() which might only be evaluated at runtime:
        if self._setupfunc is not None:
            try:
                if asyncio.iscoroutinefunction(self._setupfunc):
                    setupvars = local_loop.run_until_complete(
                        self._setupfunc(*self.args, **self.kwargs)
                    )
                else:
                    setupvars = self._setupfunc(*self.args, **self.kwargs)

            # If _setupfunc raises FailedSetup with a backup value,
            # use it as the field's new constant_output and update the
            # bar buffer:
            except FailedSetup as e:
                backup = e.args[0]
                contents = self._format_contents(
                    backup,
                    self.icon,
                    self.template,
                    self.always_show_icon
                )
                self.constant_output = contents
                bar._buffers[self.name] = str(contents)
                return

            # On success, give new values to kwargs to pass to func().
            self.kwargs['setupvars'] = setupvars

        # Run at least once at the start to ensure bar is not empty:
        if self.is_async:
            result = local_loop.run_until_complete(
                func(*self.args, **self.kwargs)
            )
        else:
            result = func(*self.args, **self.kwargs)
        last_val = result
        contents = self._format_contents(
            result,
            self.icon,
            self.template,
            self.always_show_icon
        )
        bar._buffers[self.name] = contents

        if self.run_once or once:
            local_loop.stop()
            local_loop.close()
            bar._threads.pop(self.name)  # Eventually, id(self)
            return

        step = bar._thread_cooldown
        needed = round(self.interval / step)
        count = 0
        first_cycle = True

        if self.align_to_seconds:
            # Sleep until the beginning of the next second.
            clock = time.time
            time.sleep(1 - (clock() % 1))

        start_time = clock()
        while running():

            if bar.count:
                # Stop before running a cycle that could last too long:
                if (
                    bar._print_countdown == 0
                    or self.interval > (
                        bar._print_countdown * bar.refresh_rate 
                    )
                ):
                    local_loop.stop()
                    local_loop.close()
                    bar._coros.pop(self.name, None)
                    return


            # Sleep until the next refresh cycle in little steps to
            # check if the bar has stopped.
            if count < needed and not first_cycle:
                count += 1

                # Latency from nonzero execution times causes drift,
                # where sleeps become out-of-sync and the bar skips
                # field updates.
                # This is especially noticeable in fields that update
                # routinely, such as the time.

                # To negate drift, when sleeping until the beginning of
                # the next cycle, we must also compensate for latency.
                time.sleep(
                    step - (
                        # Get the current latency, which can vary:
                        clock() % step
                        # (clock() - start_time) % step To preserve offset
                    )
                )
                continue

            count = 0
            if first_cycle:
                first_cycle = False

            if self.is_async:
                result = local_loop.run_until_complete(
                    func(*self.args, **self.kwargs)
                )
            else:
                result = func(*self.args, **self.kwargs)

            if result == last_val:
                continue
            last_val = result

            if using_format_str:
                contents = self.template.format(result, icon=self.icon)
            else:
                if self.always_show_icon or result:
                    contents = self.icon + result
                else:
                    contents = result

            bar._buffers[self.name] = contents

            # Send new field contents to the bar's override queue and print a
            # new line between refresh cycles.
            if self.overrides_refresh:
                try:
                    bar._override_queue._loop.call_soon_threadsafe(
                        bar._override_queue.put_nowait, (self.name, contents)
                    )

                except asyncio.QueueFull:
                    # Since the bar buffer was just updated, do nothing if the
                    # queue is full. The update may still show while the queue
                    # handles the current override.
                    # If not, the line will update at the next refresh cycle.
                    pass

        local_loop.stop()
        local_loop.close()

    @staticmethod
    def _check_bar(
        bar: Bar_T,
        raise_on_fail: bool = True
    ) -> NoReturn | bool:
        '''
        Check if a bar has the right attributes to use a run() function.
        If these attributes are missing,
            return ``False`` if `raise_on_fail` is ``True``,
            or raise :exc:`InvalidBarError` if `raise_on_fail` is ``False``.

        :param bar: The status bar object to test
        :type bar: :class:`Bar`

        :param raise_on_fail: Raise an exception if `bar` fails the check,
            defaults to ``False``
        :type raise_on_fail: :class:`bool`

        :raises: :exc:`InvalidBarError` if the stream fails the test
            and lacks required attributes
        '''
        required_attrs = (
            '_buffers',
            '_can_run',
            '_override_queue',
            '_stream',
        )
        if not all(hasattr(bar, a) for a in required_attrs):
            if not raise_on_fail:
                return False
            raise InvalidBarError(
                f"Status bar {bar!r} is missing certain attributes"
                f" required to run `Field.run()`"
                f" and `Field.run_threaded()`."
            ) from None
        return True

    def make_thread(self, bar: Bar_T, run_once: bool) -> None:
        '''
        Return a thread that runs the :class:`Field`'s callback.

        :param bar: Send results to this status bar
        :type bar: :class:`Bar`

        :param once: Run the Field function once
        :type once: :class:`bool`
        '''
        thread = threading.Thread(
            target=self.run_threaded,
            args=(bar, run_once),
            name=self.name  # #NOTE Eventually, id(self)
        )
        return thread


FieldPrecursor: TypeAlias = FieldName | Field | FormatterFieldSig
from . import _types
_types.FieldPrecursor = FieldPrecursor




from .bar import Bar, BarTemplate
import asyncio
import json
import os
import sys
import threading
import time
from copy import deepcopy
from string import Formatter

from .constants import CONFIG_FILE, DEBUG
from . import cli
from . import utils
from .errors import *
from .field import Field, FieldPrecursor
from .formatting import FormatterFieldSig
from ._types import (
    Args,
    BarSpec,
    BarTemplateSpec,
    ConsoleControlCode,
    FieldName,
    FieldOrder,
    FmtStrStructure,
    FormatStr,
    Icon,
    JSONText,
    Kwargs,
    Line,
    PTY_Icon,
    PTY_Separator,
    Pattern,
    Separator,
    TTY_Icon,
    TTY_Separator,
)

from collections.abc import Iterable, Iterator, Sequence
from os import PathLike
from typing import IO, NoReturn, Required, Self, TypeAlias, TypedDict, TypeVar


# Unix terminal escape code (control sequence introducer):
CSI: ConsoleControlCode = '\033['
CLEAR_LINE: ConsoleControlCode = '\x1b[2K'  # VT100 escape code to clear line
HIDE_CURSOR: ConsoleControlCode = '?25l'
UNHIDE_CURSOR: ConsoleControlCode = '?25h'



class SyncField(Field):
    # For threaded fields, still run continuously!!!

    def _do_setup(self,) : #-> SetupVars|Content?  # side-effects?

        # Use the pre-defined _setupfunc() to gather constant variables
        # for func() which might only be evaluated at runtime:
        if self._setupfunc is not None:
            try:
##                if asyncio.iscoroutinefunction(self._setupfunc):
##                    setupvars = (
##                        await self._setupfunc(*self.args, **self.kwargs)
##                    )
##                else:
                setupvars = self._setupfunc(*self.args, **self.kwargs)

            # If _setupfunc raises FailedSetup with a backup value,
            # use it as the field's new constant_output:
            except FailedSetup as e:
                backup = e.backup
                contents = self._auto_format(backup)
                self.constant_output = contents  # strcontents() everywhere?
##                return contents

            # On success, give new values to kwargs to pass to func().
            return setupvars
            self.kwargs['setupvars'] = setupvars  # Do we want side-effects???


    def _auto_format(self, text: str) -> 'Contents':
        '''Non-staticmethod _format_contents...'''
        if self.fmt is None:
            if self.always_show_icon or text:
                return self.icon + text
            else:
                return text
        else:
            return self.fmt.format(text, icon=self.icon)

    def run(self, once: bool = None) -> 'Contents':
        # hook into setup() beforehand
        if asyncio.iscoroutinefunction(self._func):
            return asyncio.run(self._func(*self.args, **self.kwargs))
        return self._func(*self.args, **self.kwargs)

        # Do not run fields which have a constant output;
        # only set their bar buffer.
        if self.constant_output is not None:
            return self._auto_contents(self.constant_output)

        result = _func(*self.args, **self.kwargs)
        contents = self._auto_format(result)

        return contents


class SyncBar(Bar):
    @classmethod
    def from_template(
        cls,
        tmpl: BarTemplate,
        *,
        ignore_with: Pattern | tuple[Pattern] | None = '//'
    ) -> Self:
        '''
        Return a new :class:`Bar` from a :class:`BarTemplate`
        or a dict of :class:`Bar` parameters.
        Ignore keys and list elements starting with `ignore_with`,
        which is ``'//'`` by default.
        If :param ignore_with: is ``None``, do not remove any values.

        :param tmpl: The :class:`dict` to convert
        :type tmpl: :class:`dict`

        :param ignore_with: A pattern to ignore, defaults to ``'//'``
        :type ignore_with: :class:`_types.Pattern` | tuple[:class:`_types.Pattern`] | ``None``, optional

        :returns: A new :class:`Bar`
        :rtype: :class:`Bar`

        :raises: :exc:`errors.IncompatibleArgsError` when
            no `field_order` is defined
            or when no `fmt` is defined
        :raises: :exc:`errors.DefaultFieldNotFoundError` when either
            a field in `field_order` or
            a field in `fmt`
            cannot be found in :attr:`Field._default_fields`
        :raises: :exc:`errors.UndefinedFieldError` when
            a custom field name in `field_order` or
            a custom field name in `fmt`
            is not properly defined in `field_definitions`
        :raises: :exc:`errors.InvalidFieldSpecError` when
            a field definition is not of the form :class:`_types.FieldSpec`

        .. note:: `tmpl` can be a regular :class:`dict`
        as long as it matches the form of :class:`_types.BarSpec`.

        '''


##        if SYNC:
##            Field = SyncField


        if ignore_with is None:
            data = deepcopy(tmpl)
        else:
            data = utils.scrub_comments(tmpl, ignore_with)

        bar_params = cls._default_params | data
        field_order = bar_params.pop('field_order', None)
        field_icons = bar_params.pop('field_icons', {})  # From the CLI
        field_defs = bar_params.pop('field_definitions', {})

        if (fmt := bar_params.get('fmt')) is None:
            if field_order is None:
                raise IncompatibleArgsError(
                    "A bar format string 'fmt' is required when field "
                    "order list 'field_order' is undefined."
                )
        else:
            field_order, field_sigs = cls.parse_fmt(fmt)

        # Ensure icon assignments correspond to valid fields:
        for name in field_icons:
            if name not in field_order:
                deduped = dict.fromkeys(field_order)  # Preserve order
                expctd_from_icons = utils.join_options(deduped)
                exc = utils.make_error_message(
                    InvalidFieldError,
                    doing_what="parsing custom Field icons",
                    blame=f"{name!r}",
                    expected=f"a Field name from among {expctd_from_icons}",
                    epilogue="Only assign icons to Fields that will be in the Bar."
                )
                raise exc from None

        # Gather Field parameters and instantiate them:
        fields = {}
        expctd_name="the name of a default or properly defined `custom` Field"

        for name in field_order:
            field_params = field_defs.get(name)

            match field_params:
                case None:
                    # The field is strictly default.
                    if name in fields:
                        continue

                    try:
                        field = SyncField.from_default(name)
                    except DefaultFieldNotFoundError:
                        exc = utils.make_error_message(
                            DefaultFieldNotFoundError,
                            # doing_what="parsing 'field_order'",  # Only relevant to config file parsing
                            blame=f"{name!r}",
                            expected=expctd_name
                        )
                        raise exc from None

                case {'custom': True}:
                    # The field is custom, so it is only defined in
                    # 'field_definitions' and will not be found as a default.
                    name = field_params.pop('name', name)
                    del field_params['custom']
                    field = SyncField(**field_params, name=name)

                case {}:
                    # The field is a default overridden by the user.
                    # Are there custom icons given from the CLI?
                    if name in field_icons:
                        cust_icon = field_icons.pop(name)
                        field_params['icons'] = (cust_icon, cust_icon)

                    try:
                        field = SyncField.from_default(name, overrides=field_params)
                    except DefaultFieldNotFoundError:
                        exc = utils.make_error_message(
                            UndefinedFieldError,
                            # doing_what="parsing 'field_order'",
                            blame=f"{name!r}",
                            expected=expctd_name,
                            epilogue=(
                                f"(In config files, remember to set "
                                f"`custom=true` "
                                f"for custom field definitions.)"
                            ),
                        )
                        raise exc from None

                case _:
                    # Edge cases.
                    exc = utils.make_error_message(
                        InvalidFieldSpecError,
                        # doing_what="parsing 'field_definitions'",
                        doing_what=f"parsing {name!r} definition",
                        details=(
                            f"Invalid Field specification: {field_params!r}",
                        ),
                        indent_level=1
                    )
                    raise exc from None

            fields[name] = field

        bar = cls(
            fields=tuple(fields.values()),
            field_order=field_order,
            **bar_params
        )
        return bar

    def run(self, once: bool = None, stream: IO = None) -> None:
        '''
        Run the bar in the specified output stream.
        Block until an exception is raised and exit smoothly.

        :param stream: The IO stream in which to run the bar,
            defaults to :attr:`Bar._stream`
        :type stream: :class:`IO`

        :param once: Whether to print the bar only once, defaults to ``False``
        :type once: :class:`bool`

        :returns: ``None``
        '''
        if stream is not None:
            self._stream = stream
        if once is None:
            once = self.run_once

        try:
            self._can_run.set()
            overriding = False

            for field in self:
                if field.overrides_refresh:
                    overriding = True

                if field.threaded:
                    field.sync_send_to_thread(run_once=once)

            if once:
                for field in self:
                    field.run()
                # Wait for threads to finish:
                while self._threads:
                    time.sleep(self._thread_cooldown)
                self._print_one_line()

            else:
                if overriding:
                    self._handle_overrides()
                self._continuous_line_printer()

        except KeyboardInterrupt:
            pass

        finally:
            self._shutdown()


    def _continuous_line_printer(self, end: str = '\r') -> None:
        using_format_str = (self.fmt is not None)
        sep = self.separator
        running = self._can_run.is_set
        clock = time.monotonic

        if self.in_a_tty:
            beginning = self.clearline_char + end
            self._stream.write(CSI + HIDE_CURSOR)
        else:
            beginning = self.clearline_char

        # fields = set(self._fields) - self._threads
        fields = tuple(f for f in self if not f.threaded)
        print(fields)

        # Flushing the buffer before writing to it fixes poor i3bar alignment.
        self._stream.flush()

##        # Print something right away just so that the bar is not empty:
##        self._print_one_line()

        if self.align_to_seconds:
            # Begin every refresh at the start of a clock second:
            clock = time.time
            time.sleep(1 - (clock() % 1))

        start_time = clock()
        for f in fields:
            print(f._func)
        while running():
            self._buffers.update((f.name, f.run()) for f in fields)
                
            if using_format_str:
                line = self.fmt.format_map(self._buffers)
            else:
                line = sep.join(
                    buf for field in fields
                        if (buf := self._buffers[field.name])
                        or self.join_empty_fields
                )

            self._stream.write(beginning + line + end)
            self._stream.flush()

            # Sleep only until the next possible refresh to keep the
            # refresh cycle length consistent and prevent drifting.
            time.sleep(
                # Time until next refresh:
                self.refresh_rate - (
                    # Get the current latency, which can vary:
                    (clock() - start_time) % self.refresh_rate  # Preserve offset
                )
            )

        if self.in_a_tty:
            self._stream.write('\n')
            self._stream.write(CSI + UNHIDE_CURSOR)

    def line_generator(self):
        using_format_str = (self.fmt is not None)
        sep = self.separator
        running = self._can_run.is_set

        while running():
            if using_format_str:
                line = self.fmt.format_map(self._buffers)
            else:
                line = sep.join(
                    buf for field in self._field_order
                        if (buf := self._buffers[field])
                        or self.join_empty_fields
                )
            yield line






import os.path
import sys
from typing import NoReturn

from .bar import Bar
from . import constants


def main() -> NoReturn | None:
    '''Run the command line utility.'''
    try:
        bar = Bar.from_cli()
        bar.run()
    except KeyboardInterrupt:
        print()
        sys.exit(1)


if __name__ == '__main__':
    main()




__all__ = (
    'CONFIG_FILE',
    'DEBUG',
    'CSI',
    'CLEAR_LINE',
    'HIDE_CURSOR',
    'UNHIDE_CURSOR',
    'FONTAWESOME_ICONS',
    'USING_FONTAWESOME'
)


from ._types import ConsoleControlCode
import os.path


CONFIG_FILE: str = os.path.abspath(os.path.expanduser(
    '~/.config/mybar/conf.json'
))
'''The default bar config file path.'''

DEBUG: bool = False
'''The default debug state.'''


# Used by Bar:
CSI: ConsoleControlCode = '\033['
'''Unix terminal escape code (control sequence introducer).'''

CLEAR_LINE: ConsoleControlCode = '\x1b[2K'
'''VT100 escape code to clear the line.'''

HIDE_CURSOR: ConsoleControlCode = '?25l'
'''VT100 escape code to hide the cursor.'''

UNHIDE_CURSOR: ConsoleControlCode = '?25h'
'''VT100 escape code to unhide the cursor.'''


FONTAWESOME_ICONS = {
    'uptime': '',
    'cpu_usage': '',
    'cpu_temp': '',
    'mem_usage': '',
    'disk_usage': '',
    'battery': '',
    'net_stats': '',
}

USING_FONTAWESOME = False




"""
######
mybar
######

*Craft highly customizable status bars with ease.*

**mybar** is a versatile status bar library and tool written in Python.


:copyright: (c) 2021-present by LonelyAbsol.
:license: MIT, see LICENSE for more details.
"""

__title__ = 'mybar'
__description__ = "An async status bar with a highly customizable API."
__url__ = "https://github.com/lonelyabsol/mybar"
__version__ = '0.8'
__author__ = "LonelyAbsol"
__license__ = 'MIT'
__copyright__ = "Copyright (c) 2021-present LonelyAbsol"


from .constants import *
from . import utils
from .bar import *
from .field import *

